/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * Alignment.cpp
 */

#include "Alignment.h"
#include "ShuffledAlignment.h"
#include "Tree.h"
#include "Utilities.h"
#include "LikelihoodCache.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <fstream>
#include <iostream>
#include <sstream>
#include <stack>
#include <string>
#include <stdexcept>

const string Alignment::TCOFFEE_LOC = "t_coffee";

int Alignment::minStemLength = 3;
int Alignment::minLoopLength = 2;

Alignment::Alignment() {
	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

Alignment::Alignment(vector<string> names, vector<string> alignment, string codon){
	//copy names and alignments
	seqNames = names;
	alignedSeqs = alignment;

	codonAnnotation = codon;
	size_t match = codonAnnotation.find_first_of("123");
	annotated = (match != string::npos);

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

Alignment::Alignment(string & filename) {
	readAlignment(filename);

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}


Alignment::Alignment(string & filename, string & annotation_file)
{
	readAlignment(filename);

	ifstream codons(annotation_file.c_str());
	string line;
	string seq;

	// Read in codon annotation
	codonAnnotation.clear();
	if(!codons.is_open()){
		cerr << "Error: cannot open codon annotation file " << annotation_file << endl;
		exit(-1);
	}
	while(codons.good()){
		getline(codons, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			// Ignore header line
		}
		else{
			seq = seq + line;
		}
	}

	// Check codon annotation validity
	if (seq.length() == 0) {
		cerr << "Error: No valid data found in " << annotation_file << endl;
		exit(-1);
	}
	size_t match = seq.find_first_of("0123");
	if (match != string::npos){
		// Check validity of codon annotation
		if(seq.length() != this->length) {
			cerr << "Error: Codon annotation length does not match that of first sequence" << endl;
			exit(-1);
		}
		size_t match = seq.find_first_not_of("0123");
		if (match != string::npos) {
			cerr << "Error: Invalid codon annotation '" << seq[match] << "' at position " << match << endl;
			exit(-1);
		}
		codonAnnotation = seq;
	} else {
		cerr << "Error: No valid codon annotation found in " << annotation_file << endl;
		exit(-1);
	}

	match = codonAnnotation.find_first_of("123");
	annotated = (match != string::npos) ? true : false;

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

Alignment::Alignment(Alignment* a, Alignment* b) {

	assert(a->seqNames.size() == b->seqNames.size());

	for (unsigned int i = 0 ; i < a->seqNames.size() ; i++) {
		seqNames = a->seqNames;
		alignedSeqs.push_back(a->alignedSeqs[i] + b->alignedSeqs[i]);
	}

	// Combine codon annotations
	codonAnnotation = a->codonAnnotation;
	codonAnnotation.insert(codonAnnotation.end(), b->codonAnnotation.begin(), b->codonAnnotation.end());
	annotated = a->annotated || b->annotated;

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();
	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

//extract unaligned seqs and make map of unaligned seq positions to aligned seq positions
void Alignment::initializeAlignmentMap() {
	vector<int>* temp;
	vector<int>* temp2;
	string seq;

	unsigned int alignedSeqLength = alignedSeqs[0].length();

	for (unsigned int i = 0; i < alignedSeqs.size(); i++){

		assert(alignedSeqLength == alignedSeqs[i].length()); //sanity check
		temp = new vector<int>();
		temp2 = new vector<int>();
		seq = "";

		for (unsigned int j = 0; j < alignedSeqLength; j++){
			if (alignedSeqs[i].at(j) != '-'){
				temp->push_back(j);
				seq += alignedSeqs[i].at(j);
				temp2->push_back(temp->size() - 1);
			}
			else{
				temp2->push_back(-1);
			}
		}
		seqs.push_back(seq);
		seq2AlignmentMap.push_back(temp);
		alignment2SeqMap.push_back(temp2);
	}
}


/*
 * This creates the likelihood cache object
 * You must call this method on the Alignment object if you wish to
 * calculate log likelihood values.
 */
void Alignment::initializeCache() {
	hasCache = true;
	likelihoods = new LikelihoodCache(this->length);
}

Alignment::~Alignment() {
	size_t i;
	for (i = 0; i < seq2AlignmentMap.size(); i++){
		delete seq2AlignmentMap[i];
	}
	seq2AlignmentMap.clear();

	for (i = 0; i < alignment2SeqMap.size(); i++){
		delete alignment2SeqMap[i];
	}
	alignment2SeqMap.clear();

	if (hasCache) {
		delete likelihoods;
		likelihoods = NULL;
	}

	delete consensusBP;
	consensusBP = NULL;

	delete seqCons;
	seqCons = NULL;
}

string Alignment::seqString()
{
	string out = "";
	for(unsigned int i = 0; i < seqs.size(); i++){
		out += seqs[i];
		out += "\n";
	}
	return out;
}

string Alignment::columnString(int pos)
{
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		//cout << *(alignedSeqs[i]) << endl;
		out += alignedSeqs[i].at(pos);
	}
	return out;
}


string Alignment::alignmentString()
{
	// cout << "Alignment Size: " << alignedSeqs.size() << endl;
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		//cout << *(alignedSeqs[i]) << endl;
		out += alignedSeqs[i];
		out += "\n";
	}
	return out;
}

string Alignment::codonString()
{
	return codonAnnotation;
}

void Alignment::readAlignment(string & filename)
{
	ifstream fastaAln(filename.c_str());
	string line;
	string seq = "";
	string name = "";
	if(!fastaAln.is_open()){
		cerr << "Error: cannot open alignment file " << filename << endl;
		exit(-1);
	}
	while(fastaAln.good()){
		getline(fastaAln, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			// Write previous entry
			if(seq != ""){
				//convert to upper case, and convert Ts to Us
				for(unsigned int i = 0; i < seq.length(); i++){
					seq[i] = toupper(seq[i]);
					if(seq[i] == 'T'){
						seq[i] = 'U';
					}
					if(seq[i] == '.') {
						seq[i] = '-';
					}
				}
				alignedSeqs.push_back(seq);
			}
			// Start new entry
			seqNames.push_back(line.substr(1));
			seq = "";
		}
		else{
			seq = seq + line;
		}
	}

	// Process last line
	assert(seq.length() > 0);

	size_t match = seq.find_first_of("0123");
	if (match != string::npos){
		//last sequence contains the codon annotation

		// Check validity of codon annotation
		if(seq.length() != alignedSeqs[0].size()) {
			cerr << "Error: Codon annotation length does not match that of first sequence" << endl;
			exit(-1);
		}
		size_t match = seq.find_first_not_of("0123");
		if (match != string::npos) {
			cerr << "Error: Invalid codon annotation '" << seq[match] << "' at position " << match << endl;
			exit(-1);
		}

		seqNames.pop_back(); // remove last species name (corresponding to codon annotation)
		codonAnnotation = seq;

		match = seq.find_first_of("123");
		annotated = (match != string::npos) ? true : false;
	}
	else{
		annotated = false;
		//convert to upper case, and convert Ts to Us
		for(unsigned int i = 0; i < seq.length(); i++){
			seq[i] = toupper(seq[i]);
			if(seq[i] == 'T'){
				seq[i] = 'U';
			}
			codonAnnotation += "0";
		}
		alignedSeqs.push_back(seq);
	}

	fastaAln.close();

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();
}


/*
 * The following three functions are used to retrieve likelihoods from the alignment
 * Precondition: you must first call initializeCache() before calling any of these
 * methods. This saves space for alignments which are not used to calculate likelihoods,
 * as they do not require cache memory
 */
double Alignment::logLikelihood(int pos5, int pos3, Tree * tree){
	return this->logPairedLikelihood(pos5, pos3, tree)
			- this->logUnpairedLikelihood(pos5, tree)
			- this->logUnpairedLikelihood(pos3, tree);
}

double Alignment::logPairedLikelihood(int pos5, int pos3, Tree * tree){
	if (hasCache && likelihoods->isCached(pos5, pos3)) {
		return likelihoods->getCached(pos5, pos3);
	}
	else {
		double likelihood = tree->calcLogFelsPaired(*this, pos5, pos3);
		if (hasCache) likelihoods->setCached(pos5, pos3, likelihood);
		return likelihood;
	}
}

double Alignment::logUnpairedLikelihood(int pos, Tree * tree){
	if (hasCache && likelihoods->isCached(pos)) {
		return likelihoods->getCached(pos);
	}
	else {
		double likelihood = tree->calcLogFelsUnpaired(*this, pos);
		if (hasCache) likelihoods->setCached(pos, likelihood);
		return likelihood;
	}
}


/*
 * This returns a pointer to a new Alignment object that has been realigned by TCoffee
 * Note: You must delete this object once it is not needed any more.
 */
Alignment* Alignment::realignTcoffee(Tree * tree){
	//write fasta file:
	string fastaFilenameTemplate = "/tmp/fasta.XXXXXX";
	char fastaFilename[fastaFilenameTemplate.size()+1];
	strcpy(fastaFilename, fastaFilenameTemplate.c_str());

	int unique_fd;
	FILE * unique_file;

	if ((unique_fd = mkstemp(fastaFilename)) == -1 ){
		cerr << "Could not create temporary clustal file for shuffling\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary clustal file for shuffling\n";
		exit(-1);
	}

	fputs(FastaFormat(false).c_str(), unique_file);
	fclose(unique_file);

	//make a temp file for output:
	string outTemplate = "/tmp/realign_out.XXXXXX";
	char outFilename[outTemplate.size()+1];
	strcpy(outFilename, outTemplate.c_str());

	if ((unique_fd = mkstemp(outFilename)) == -1 ){
		cerr << "Could not create temporary out file for realignment\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary out file for realignment\n";
		exit(-1);
	}
	string temp_contents = "TEMP!\n"; // this will be replaced by TCOFFEE output
	fputs(temp_contents.c_str(), unique_file);
	fclose(unique_file);

	//make a temp file for output:
	string treeTemplate = "/tmp/tree_out.XXXXXX";
	char treeFilename[treeTemplate.size()+1];
	strcpy(treeFilename, treeTemplate.c_str());

	if ((unique_fd = mkstemp(treeFilename)) == -1 ){
		cerr << "Could not create temporary tree file for realignment\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary out file for realignment\n";
		exit(-1);
	}
	//write tree to file
	fputs(tree->newickString().c_str(), unique_file);
	fclose(unique_file);

	//run realignment:
	string command = TCOFFEE_LOC + " " + string(fastaFilename) + " -outfile " + string(outFilename) + " -usetree " + string(treeFilename) + " -output fasta -n_core 1 -outorder input -quiet";
	if(system(command.c_str()) != 0){
		cerr << "Error: t_coffee encountered an error\n";
		exit(-1);
	}

	ifstream realignFile(outFilename);

	if(!realignFile.is_open()){
		cerr << "Error: cannot open realignment fasta file " << string(outFilename) << endl;
		exit(-1);
	}

	string line;
	vector<string> realignment;
	vector<string> r_names;
	string seq;
	bool found_seq = false;

	while(realignFile.good()){
		getline(realignFile, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			if(found_seq){
				//cout << "Sequence Found: " << *seq << endl;
				realignment.push_back(seq);
			}
			line = line.substr(1);
			r_names.push_back(line);
			found_seq = true;
			seq = "";
		}
		else{
			seq = seq + line;
		}
	}

	realignFile.close();
	realignment.push_back(seq);

	assert(realignment.size() == r_names.size());
	assert(realignment.size() == alignedSeqs.size());

	for(unsigned int j = 0; j < r_names.size(); j++){
		assert(r_names[j].compare(seqNames[j]) == 0);
		assert(realignment[j].length() == realignment[0].length());
	}
	Alignment* resultAlignment = new Alignment(this->seqNames, realignment, this->codonAnnotation);
	resultAlignment->realignCodonAnnotation(this->alignedSeqs[0]);

	//clean up files
	command = "rm -f " + string(outFilename) + " " + string(fastaFilename) + " " + string(treeFilename) ;
	if(system(command.c_str()) != 0){
		cerr << "Error cleaning up temporary shuffle output file\n";
		exit(-1);
	}

	return resultAlignment;
}

double Alignment::canonicalBPPercent(int pos5, int pos3){

	if(consensusBP == NULL){
		consensusBP = new UTMatrix(this->length);

		for(unsigned int i = 0; i < this->length; i++){
			for(unsigned int j = i; j < this->length; j++){
				consensusBP->set(i, j, -1);
			}
		}
	}

	if(consensusBP->at(pos5, pos3) < 0){
		int total = 0;
		int cbp = 0;
		double cbp_percent;

		for(unsigned int i = 0; i < alignedSeqs.size(); i++){
			char a,b;
			a = alignedSeqs[i].at(pos5);
			b = alignedSeqs[i].at(pos3);

			total++;
			if(Utilities::validBP(a, b)){
				cbp++;
			}
		}
		assert(total > 0);
		cbp_percent = cbp / (double) total;
		consensusBP->set(pos5, pos3, cbp_percent);
		return cbp_percent;
	}
	else{
		return consensusBP->at(pos5, pos3);
	}

}

double Alignment::getSeqCons(unsigned int pos){
	if (seqCons == NULL){
		seqCons = new vector<double>();
		for(unsigned int i = 0; i < this->length; i++){
			seqCons->push_back(-1);
		}
	}
	if(seqCons->at(pos) < 0){

		int total = 0;
		int matching = 0;
		for(unsigned int i = 0; i < alignedSeqs.size() - 1; i++){
			for(unsigned int j = i + 1; j < alignedSeqs.size(); j++){
				if(toupper(alignedSeqs[i].at(pos)) == toupper(alignedSeqs[j].at(pos))){
					matching++;
				}
				total++;
			}
		}
		(*seqCons)[pos] = matching / (double) total;
		//(*seqCons)[pos] = seqConservationAtPosition(pos);
	}
	return seqCons->at(pos);

}


double Alignment::getCovariance(int pos5, int pos3){
	int denom = 0;
	int covariance = 0;
	for(unsigned int i = 0; i < alignedSeqs.size() - 1; i++){

		for(unsigned int j = i+1; j < alignedSeqs.size(); j++){
			int hamming = 0;
			char ia, ib, ja, jb;
			ia = toupper(alignedSeqs[i].at(pos5));
			ib = toupper(alignedSeqs[i].at(pos3));

			ja = toupper(alignedSeqs[j].at(pos5));
			jb = toupper(alignedSeqs[j].at(pos3));
			if(ia != ja){
				hamming++;
			}
			if(ib != jb){
				hamming++;
			}

			if(hamming > 0){
				if(Utilities::validBP(ia, ib) && Utilities::validBP(ja,jb)){
					covariance += hamming;
				}
				else{
					covariance -= hamming;
				}
			}
			denom++;
		}
	}
	return covariance / (double)denom;
}

string Alignment::FastaFormat(bool aligned) const{
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		out += ">" + seqNames[i] + "\n";
		if(aligned){
			out += alignedSeqs[i] + "\n";
		}
		else{
			out += seqs[i] + "\n";
		}
	}
	return out;
}

// Realigns the old codon annotation given a new sequence
// Precondition: old_seq is the first sequence in the alignment prior to realignment
void Alignment::realignCodonAnnotation(const string & old_seq) {
	// cout << old_seq << endl;
	// cout << old_cod << endl;
	// cout << new_seq << endl;

	string old_cod = this->codonAnnotation;
	string new_seq = this->alignedSeqs[0];

	assert(old_seq.length() == old_cod.length());
	int new_pos = 0;
	string new_cod = "";
	char addition = (old_cod[0] == '0') ? '0' : '3';
	bool keep = false;
	for(size_t i = 0; i < old_seq.length(); ++i) {

		if(old_seq[i] == new_seq[new_pos]) {
			new_cod += old_cod[i];
			++new_pos;
			keep = false;
		} else if (new_seq[new_pos] == '-') {
			new_cod += addition;
			++new_pos;
			--i;
		} // old_seq[i] == '-', do nothing

		if(i == old_cod.length() - 1) {
			addition = (old_cod[i] == '0') ? '0' : '3';
		} else {
			// cout << i << "(" << old_cod[i] << ", " << old_cod[i + 1] << "), ";
			if (old_cod[i] == '0' && old_cod[i + 1] == '0') {
				addition = '0';
			} else if (old_cod[i] != '0' && old_cod[i + 1] != '0') {
				addition = '3';
			} else if (!keep){
				addition = (rand() % 2) ? '0' : '3';
				keep = true;
			}
		}
	}
	new_cod += string(new_seq.length() - new_pos, addition);
	this->codonAnnotation = new_cod;
}
