/*
 *  Implementation of the tree class
 *  This is a template class which takes as template parameter the data type for storing probabilities in the felsenstein calculation
 *  The two values for this parameter set in TransatMain are double and BigDouble
 *
 *  Only TransatMain accesses this class directly. All other classes use its interface: Tree
 *  This allows the probability data type to be abstracted from all uses of the class
 *
 *  Class BigDouble was used to solve issues with floating point underflow. See this class for more info on this. To use
 *  BigDouble in the calculation, set command-line parameter '-precision'
 *			-jeff
 */
#ifndef TREEIMPL_H_
#define TREEIMPL_H_

#include <list>
#include <string>
#include <vector>
#include <typeinfo>


#include <cassert>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <limits>
#include <sstream>

#include "Alignment.h"
#include "EvolModel.h"
#include "Utilities.h"
#include "BigDouble.h"


class Tree;

// subclass with specified probability data type
template <class prob_type>
class TreeImpl : public Tree{
public:

	TreeImpl<prob_type>* leftChild;
	TreeImpl<prob_type>* rightChild;
	TreeImpl<prob_type>* parent;
	double branchLength;
	string seqName; //for leaves only - otherwise = "";

	// For Felsenstein likelihood calculations
	prob_type fels[16];

	double matrix_unpaired[4][4][4];

	double matrix16_00[16][16];
	double matrix16_11[16][16];
	double matrix16_12[16][16];
	double matrix16_13[16][16];
	double matrix16_21[16][16];
	double matrix16_22[16][16];
	double matrix16_23[16][16];
	double matrix16_31[16][16];
	double matrix16_32[16][16];
	double matrix16_33[16][16];

	typedef double array_of_4_double_t[4];
	typedef double array_of_16_double_t[16];

	array_of_16_double_t* matrix_paired[4][4];



	TreeImpl(string filename) {
		//prob_type test;
		//cerr << "probability data type: " << typeid(test).name() << endl;

		ifstream treeFile(filename.c_str());
		string line;
		string newickString = "";

		if(!treeFile.is_open()){
			cerr << "Error: cannot open tree file " << filename << endl;
			exit(-1);
		}

		while(treeFile.good()){
			getline(treeFile, line);
			newickString += line;
		}

		this->parent = NULL;
		this->leftChild = NULL;
		this->rightChild = NULL;

		vector<string*> tokens;
		unsigned int split;
		unsigned int start = 0;
		string * token;
		while(start < newickString.length()){
			split = newickString.find_first_of("(),:;\'", start);
			if(split > start){
				token = new string(newickString.substr(start, split-start));
				tokens.push_back(token);
			}

			if(newickString.substr(split, 1).compare("\'") == 0) { // does not split tokens surrounded by single quotes
				unsigned int quoteStart = split+1;
				unsigned int quoteEnd = newickString.find_first_of("\'", quoteStart);
				token = new string(newickString.substr(quoteStart, quoteEnd-quoteStart));
				tokens.push_back(token);
				start = quoteEnd+1;
			}
			else {
				token = new string(newickString.substr(split, 1));
				tokens.push_back(token);
				start = split+1;
			}
		}
		createChildren(tokens);
		setUpMatrices();

		for(vector<string*>::iterator i = tokens.begin(); i < tokens.end(); i++){
			delete (*i);
		}
	}


	TreeImpl(vector<string*> newickStringTokens)
	{
		createChildren(newickStringTokens);
		setUpMatrices();
	}


	virtual ~TreeImpl() {
		delete leftChild;
		delete rightChild;
	};

	double calcLogFelsPaired(Alignment & a, int pos5, int pos3) {
		prob_type likelihood = this->calcFelsPaired(a, pos5, pos3);
		return log(likelihood)/log(2.0);
	}

	prob_type calcFelsPaired(Alignment & a, int pos5, int pos3){
		prob_type likelihood = 0.0;

		int codon5 = a.codonAnnotation[pos5] - '0';
		int codon3 = a.codonAnnotation[pos3] - '0';

	#ifndef _OLDMATRIX
		if (codon5 == 0 && codon3 == 0) {
			codon5 = 3;
			codon3 = 3;
		}
	#endif

		assert(codon5 >= 0 && codon5 <= 3);
		assert(codon3 >= 0 && codon3 <= 3);

		//leaf:
		if(leftChild == NULL && rightChild == NULL){
			map<Tree*,unsigned int> leaf2SeqMap = getLeaf2SeqMap(a.seqNames);
			unsigned int seqIndex = leaf2SeqMap.find(this)->second;

			char base5,base3;
			base5 = a.alignedSeqs[seqIndex].at(pos5);
			base3 = a.alignedSeqs[seqIndex].at(pos3);

			//convert Ns to gaps:
			if (base5 == 'N' || base5 == 'n'){
				base5 = '-';
			}
			if (base3 == 'N' || base3 == 'n'){
				base3 = '-';
			}

			int leftInterpreted[NUM_PAIRS];
			int rightInterpreted[NUM_PAIRS];

			if(Tree::nonGapPair && base5 == '-' && base3 != '-'){
				//case: 5' base is a gap, and the other is not
				Utilities::interpret(base3, rightInterpreted);
				Utilities::interpretNonPairingGap(rightInterpreted, leftInterpreted);
			}
			else if(Tree::nonGapPair && base3 == '-' && base5 != '-'){
				//case 3' base is a gap, and the other is not
				Utilities::interpret(base5, leftInterpreted);
				Utilities::interpretNonPairingGap(leftInterpreted, rightInterpreted);
			}
			else{
				//interpret chars
				Utilities::interpret(base5, leftInterpreted);
				Utilities::interpret(base3, rightInterpreted);

			}

			for(int i = 0; i < NUM_BASES; i++){
				for(int j = 0; j < NUM_BASES; j++){
					fels[NUM_BASES * i + j] = (double)(leftInterpreted[i] & rightInterpreted[j]);
				}
			}

		}
		//inner node
		else{
			//set fels array for children
			leftChild->calcFelsPaired(a, pos5, pos3);
			rightChild->calcFelsPaired(a, pos5, pos3);

			array_of_16_double_t* left_matrix = (leftChild->matrix_paired)[codon5][codon3];
			array_of_16_double_t* right_matrix = (rightChild->matrix_paired)[codon5][codon3];

			// then calculate the likelihood based on the likelihood of the two child nodes
			for(int i = 0; i < NUM_PAIRS; i++){
				prob_type leftProbability = 0.0;
				prob_type rightProbability = 0.0;
				for(int j = 0; j < NUM_PAIRS; j++){
					leftProbability += leftChild->fels[j] * left_matrix[i][j];
					rightProbability += rightChild->fels[j] * right_matrix[i][j];
				}
				fels[i] = leftProbability * rightProbability;
			}
		}
		// TODO: implement better means of detecting floating point underflow

		//root
		if(parent == NULL){
			double* pi = EvolModel::ePiPaired[codon5][codon3];
			for(int i = 0; i < NUM_PAIRS; i++){
				likelihood += fels[i] * pi[i];
			}
		}

		return likelihood;
	}



	double calcLogFelsUnpaired(Alignment & a, int pos) {
		prob_type likelihood = calcFelsUnpaired(a, pos);
		return log(likelihood)/log(2.0);
	}

	prob_type calcFelsUnpaired(Alignment & a, int pos){
		prob_type likelihood = 0.0;
		int codon = a.codonAnnotation[pos] - '0';

	#ifndef _OLDMATRIX
		if (codon == 0) {
			codon = 3;
		}
	#endif

		assert(codon >= 0 && codon <= 3);

		//leaf
		if(leftChild == NULL && rightChild == NULL){
			// TODO: pass fels array straight into 'interpret' function. more efficient... do not need to create array and copy over
			int interpreted[NUM_BASES];
			map<Tree*,unsigned int> leaf2SeqMap = getLeaf2SeqMap(a.seqNames);
			unsigned int seqIndex = leaf2SeqMap.find(this)->second;
			Utilities::interpret(a.alignedSeqs[seqIndex].at(pos), interpreted);

			for(int i = 0; i < NUM_BASES; i++){
				// Assigns prob of 1 for presence, 0 for absence over AUGC
				fels[i] = (double)interpreted[i];
				//note: should fels be averaged, so that it sums to 1? will have to think about this...
				//answer = no. See Felsenstein 1981, Extensions section
			}
		}

		//inner node
		else{
			//set fels array for children
			leftChild->calcFelsUnpaired(a, pos);
			rightChild->calcFelsUnpaired(a, pos);

			// then calculate the likelihood here
			for(int i = 0; i < NUM_BASES; i++){
				prob_type leftProbability = 0.0;
				prob_type rightProbability = 0.0;
				for(int j = 0; j < NUM_BASES; j++){
					leftProbability += leftChild->fels[j] * (leftChild->matrix_unpaired)[codon][i][j];
					rightProbability += rightChild->fels[j] * (rightChild->matrix_unpaired)[codon][i][j];
				}
				fels[i] = leftProbability * rightProbability;
			}
		}

		//root
		if(parent == NULL){
			double* piVals = EvolModel::ePiUnpaired[codon];
			for(int i = 0; i < NUM_BASES; i++){
				likelihood += fels[i] * piVals[i];
			}
			// assert(likelihood > 0);
		}

		return likelihood;
	}

	/*
	 * prints paired likelihood matrix and unpaired likelihood list
	 */
	void printAllLikelihoods(Alignment& alignment) {
		cout << ">Paired likelihood" << endl;
		for( int i = 0; i < alignment.length; i++) {
			for (int j = 0; j < alignment.length; j++) {
				if (j != 0)
					cout << "\t";

				if (i != j) {
					cout << this->calcFelsPaired(alignment, i, j);
				}
				else {
					cout << 0;
				}
			}
			cout << endl;
		}

		cout << ">Unpaired likelihood";
		for( int i = 0; i < alignment.length; i++) {
			cout << endl;
			cout << this->calcFelsUnpaired(alignment, i);
		}
		cout << endl;
	}


	map<Tree*, unsigned int> getLeaf2SeqMap(vector<string> & SeqNames)
	{
		map<Tree*, unsigned int> out;
		if (leftChild == NULL && rightChild == NULL){
			unsigned int i;
			for(i = 0; i < SeqNames.size(); i++){
				if(seqName.compare(SeqNames[i]) == 0){
					break;
				}
			}
			if(i >= SeqNames.size()){
				cerr << "Error: tree sequence name not in set of alignment sequences\n";
				cerr << "unrecognized name: " << seqName << endl;
				cerr << "Alignment names:\n";
				for(i = 0; i < SeqNames.size();i++){
					cerr << SeqNames[i] << endl;
				}
				exit(-1);
			}
			else{
				out.insert(pair<Tree*, unsigned int>(this, i));
			}
		}
		else{
			out = leftChild->getLeaf2SeqMap(SeqNames);
			map<Tree*, unsigned int> rightMap = rightChild->getLeaf2SeqMap(SeqNames);

			out.insert(rightMap.begin(), rightMap.end());
		}

		return out;
	}

	// returns newickString representation of the tree
	string newickString(){
		stringstream ss;
		//if leaf
		if(leftChild == NULL){
			ss << seqName << ":" << branchLength;
		}
		else{
			//internal node
			ss << "(" << leftChild->newickString() << "," << rightChild->newickString() << ")";

			if(parent != NULL){
				ss << ":" << branchLength;
			}
			else{
				//root - branch length is ignored
				ss << ";";
			}
		}
		//cout << ss.str() << endl;
		return ss.str();
	}


	double totalLength(){
		double length = branchLength;

		if(leftChild != NULL){
			length += leftChild->totalLength();
		}
		if(rightChild != NULL){
			length += rightChild->totalLength();
		}
		return length;
	}

	vector<string> getSeqNames(){
		vector<string> names;

		if(leftChild == NULL && rightChild == NULL){
			names.push_back(seqName);
		}
		else{
			vector<string> temp = leftChild->getSeqNames();
			names.insert(names.end(), temp.begin(), temp.end());

			temp = rightChild->getSeqNames();
			names.insert(names.end(), temp.begin(), temp.end());
		}
		return names;
	}





private:

	// parses branch length for current node, and passes tokens to child nodes as required
	void createChildren(vector<string*> & tokens) {
		string tags("");
		tagString(tokens, tags);

		if (tags.compare("n:n") == 0 || tags.compare("l:n") == 0) {
			// leaf
			leftChild = NULL;
			rightChild = NULL;

			seqName.assign(*tokens.front());
			double tmp = atof((tokens.back())->c_str());
			tmp *= Tree::rescaleFactor;
			this->branchLength = tmp;
//			cout << this->branchLength << ",";
			return;
		}
		else {
			string leftbrace("(");
			string rightbrace(")");
			string colon(":");
			string epsilon("0.0000001");

			int childStart = 1;
			int childEnd = tags.rfind(')');
			string tail = tags.substr(childEnd);
			if (tags[0] != '(') {
				cerr << "Tree parsing error: subtree not surrounded by brackets" << endl;
			}

			// set branch length
			double tmp;
			if (tail.compare(");") == 0) {
//				this->branchLength = 0;
				tmp = 0;
//				cout << this->branchLength << ",";
			}
			else if (tail.compare("):n") == 0 || tail.compare(")n:n") == 0 || tail.compare(")l:n") == 0) {
//				this->branchLength = atof((tokens.back())->c_str());		// last token;
				tmp = atof((tokens.back())->c_str());		// last token;
//				cout << this->branchLength << ",";
			}
			else if (tail.compare("):n;") == 0 || tail.compare(")n:n;") == 0 || tail.compare(")l:n;") == 0) {
//				this->branchLength = atof((*(tokens.rbegin()+1))->c_str());	// second to last token
				tmp = atof((*(tokens.rbegin()+1))->c_str());	// second to last token
//				cout << this->branchLength << ",";
			}
			else {
				cerr << "Tree parsing error: unrecognized syntax" << endl;
			}
			tmp *= Tree::rescaleFactor;
			this->branchLength = tmp;
//			cout << this->branchLength << ",";

			// find children
			int openBraces = 0;
			int firstComma = -1;
			int numCommas = 0;
			for(int cursor = childStart; cursor < childEnd; cursor++){
				if(tags[cursor] == '(') {
					openBraces++;
				}
				if(tags[cursor] == ')') {
					openBraces--;
				}
				if(tags[cursor] == ',' && openBraces == 0) {
					numCommas++;
					if (firstComma == -1)
						firstComma = cursor;
				}
			}
			if (openBraces != 0) {
				cerr << "Tree parsing error: unbalanced brackets" << endl;
			}
			if (numCommas == 0) {
				cerr << "Tree parsing error: unrecognized subtree syntax" << endl;
			}
			vector<string*> leftTokens(tokens.begin()+childStart, tokens.begin()+(firstComma));
			vector<string*> rightTokens(tokens.begin()+(firstComma+1), tokens.begin()+childEnd);
			if (numCommas > 1) {
				rightTokens.insert(rightTokens.begin(), &leftbrace);
				rightTokens.push_back(&rightbrace);
				rightTokens.push_back(&colon);
				rightTokens.push_back(&epsilon);
			}

			this->leftChild = new TreeImpl<prob_type>(leftTokens);
			this->leftChild->parent = this;
			this->rightChild = new TreeImpl<prob_type>(rightTokens);
			this->rightChild->parent = this;
		}
	}

	// returns token descriptor tag
	// for special characters, it simply returns that character
	// for numeric tokens, returns 'n'    (numeric)
	// for all other tokens, returns 'l'  (label)
	char tokenTag(string* token) {
		if (token->compare(",") == 0) {
			return ',';
		}
		else if (token->compare("(") == 0) {
			return '(';
		}
		else if (token->compare(")") == 0) {
			return ')';
		}
		else if (token->compare(":") == 0) {
			return ':';
		}
		else if (token->compare(";") == 0) {
			return ';';
		}
		else {
			string::iterator cursor;
			for(cursor = token->begin(); cursor != token->end(); cursor++) {
				char ch = *cursor;
				if (!isdigit(ch) && ch != '-' && ch != '.' && ch != 'E' && ch != 'e') {
					return 'l';
				}
			}
			return 'n';
		}
	}
	
	// returns sequence of token tags for list of tokens
	void tagString(vector<string*> &tokens, string& out) {
		vector<string*>::iterator cursor;
		out.clear();
		for (cursor = tokens.begin(); cursor < tokens.end(); cursor++) {
			out.push_back(tokenTag(*cursor));
		}
	}	

	// debug function. prings token list with '|' separator
	void printTokens(vector<string*> &tokens) {
		vector<string*>::iterator cursor;
		for (cursor = tokens.begin(); cursor < tokens.end(); cursor++) {
			string* token = *cursor;
			cerr << "|" << *token;
		}
		cerr << endl;
	}



	// Calculates rate matrices for this node in the evolutionary tree
	void setUpMatrices(){
		//code adapted from SimulFold.TreeNode.settingTransitions()

		//calculate unpaired matrices
		for(int codon = 0; codon < 4; codon++) {
			for(int i = 0; i < 4; i++){
				for(int j = 0; j < 4; j++){
					matrix_unpaired[codon][i][j] = 0.0;
				}
			}

			array_of_4_double_t* vVals = EvolModel::eVUnpaired[codon];
			array_of_4_double_t* wVals = EvolModel::eWUnpaired[codon];
			double* dVals = EvolModel::eDUnpaired[codon];
			double r_ns_codon = EvolModel::r_ns[codon];

			for(int k = 0; k < 4; k++){
				double exponent = exp(dVals[k] * branchLength * r_ns_codon);
				for(int i = 0; i < 4; i++){
					for(int j = 0; j < 4; j++){
						matrix_unpaired[codon][i][j] += vVals[i][k] * exponent * wVals[k][j];
					}
				}
			}
		}

		int numCodonCombinations = 10;
		int codonCombinations[10][2] = {
				{0, 0}, {1, 1}, {1, 2}, {1, 3}, {2, 1}, {2, 2}, {2, 3}, {3, 1}, {3, 2}, {3, 3}
		};

		// TODO: compiler gives a warning that this construct is not available in all compilers
		//this->matrix_paired = {
		//		{matrix16_00, matrix16_31, matrix16_32, matrix16_33},
		//		{matrix16_13, matrix16_11, matrix16_12, matrix16_13},
		//		{matrix16_23, matrix16_21, matrix16_22, matrix16_23},
		//		{matrix16_33, matrix16_31, matrix16_32, matrix16_33}
		//};
		this->matrix_paired[0][0] = matrix16_00;
		this->matrix_paired[0][1] = matrix16_31;
		this->matrix_paired[0][2] = matrix16_32;
		this->matrix_paired[0][3] = matrix16_33;
		this->matrix_paired[1][0] = matrix16_13;
		this->matrix_paired[1][1] = matrix16_11;
		this->matrix_paired[1][2] = matrix16_12;
		this->matrix_paired[1][3] = matrix16_13;
		this->matrix_paired[2][0] = matrix16_23;
		this->matrix_paired[2][1] = matrix16_21;
		this->matrix_paired[2][2] = matrix16_22;
		this->matrix_paired[2][3] = matrix16_23;
		this->matrix_paired[3][0] = matrix16_33;
		this->matrix_paired[3][1] = matrix16_31;
		this->matrix_paired[3][2] = matrix16_32;
		this->matrix_paired[3][3] = matrix16_33;

		// Calculate base paired matrices
		for (int codonI = 0; codonI < numCodonCombinations; codonI++) {
			int codon1 = codonCombinations[codonI][0];
			int codon2 = codonCombinations[codonI][1];
			double r_ns_codon1 = EvolModel::r_ns[codon1];
			double r_ns_codon2 = EvolModel::r_ns[codon2];

			array_of_16_double_t* matrix = matrix_paired[codon1][codon2];
			for(int i = 0; i < 16; i++){
				for(int j = 0; j < 16; j++){
					matrix[i][j] = 0.0;
				}
			}

			double* dVals = EvolModel::eDPaired[codon1][codon2];
			array_of_16_double_t* vVals = EvolModel::eVPaired[codon1][codon2];
			array_of_16_double_t* wVals = EvolModel::eWPaired[codon1][codon2];
			double r_p_val = (codon1 == 0 && codon2 == 0) ? 1.0: EvolModel::r_p;

			for(int k = 0; k < 16; k++){
				double exponent = exp(dVals[k] * branchLength * r_p_val * ((r_ns_codon1) + (r_ns_codon2)) / 2.0);
				for(int i = 0; i < 16; i++){
					for(int j = 0; j < 16; j++){
						matrix[i][j] += vVals[i][k] * exponent * wVals[k][j];
					}
				}
			}
		}
	}
};

#endif
